# meowpow: C/C++ implementation of Meowpow, the Meowcoin Proof of Work algorithm.
# Copyright 2019 Pawel Bylica.
# Licensed under the Apache License, Version 2.0.

from _meowpow import ffi, lib


def keccak_256(data):
    hash = lib.meowpow_keccak256(ffi.from_buffer(data), len(data))
    return ffi.unpack(hash.str, len(hash.str))


def keccak_512(data):
    hash = lib.meowpow_keccak512(ffi.from_buffer(data), len(data))
    return ffi.unpack(hash.str, len(hash.str))


def hash(epoch_number, header_hash, nonce):
    if len(header_hash) != 32:
        raise ValueError('header_hash must have length of 32')

    ctx = lib.meowpow_get_global_epoch_context(epoch_number)
    c_header_hash = ffi.new('union meowpow_hash256*')
    c_header_hash[0].str = header_hash
    result = lib.meowpow_hash(ctx, c_header_hash, nonce)
    final_hash = ffi.unpack(result.final_hash.str, len(result.final_hash.str))
    mix_hash = ffi.unpack(result.mix_hash.str, len(result.mix_hash.str))
    return final_hash, mix_hash


def verify(epoch_number, header_hash, mix_hash, nonce, boundary):
    if len(header_hash) != 32:
        raise ValueError('header_hash must have length of 32')
    if len(mix_hash) != 32:
        raise ValueError('mix_hash must have length of 32')
    if len(boundary) != 32:
        raise ValueError('boundary must have length of 32')

    ctx = lib.meowpow_get_global_epoch_context(epoch_number)
    c_header_hash = ffi.new('union meowpow_hash256*')
    c_header_hash[0].str = header_hash
    c_mix_hash = ffi.new('union meowpow_hash256*')
    c_mix_hash[0].str = mix_hash
    c_boundary = ffi.new('union meowpow_hash256*')
    c_boundary[0].str = boundary

    return lib.meowpow_verify(ctx, c_header_hash, c_mix_hash, nonce, c_boundary)

def light_verify(header_hash, mix_hash, nonce):
    if len(header_hash) != 32:
        raise ValueError('header_hash must have length of 32')

    c_header_hash = ffi.new('union meowpow_hash256*')
    c_header_hash[0].str = header_hash

    c_mix_hash = ffi.new('union meowpow_hash256*')
    c_mix_hash[0].str = mix_hash

    result = lib.light_verify(c_header_hash, c_mix_hash, nonce)
    final_hash = ffi.unpack(result.str, len(result.str))
    return final_hash



#!/usr/bin/env python3

# meowpow: C/C++ implementation of Meowpow, the Meowcoin Proof of Work algorithm.
# Copyright 2019 Pawel Bylica.
# Licensed under the Apache License, Version 2.0.

import os
import subprocess
import shutil
from os import path

from setuptools import setup
from setuptools.command.build_ext import build_ext as setuptools_build_ext

try:
    from distutils.errors import CCompilerError
except Exception:  # Python 3.12+ may not have distutils
    try:
        from setuptools.errors import CompileError as CCompilerError
    except Exception:
        class CCompilerError(RuntimeError):
            pass


class build_ext(setuptools_build_ext):
    user_options = setuptools_build_ext.user_options + [
        ('skip-cmake-build', None,
         "Skip CMake build assuming the libraries are already installed " +
         "to the dist directory")
    ]

    def initialize_options(self):
        super(build_ext, self).initialize_options()
        self.skip_cmake_build = False

    def run(self):
        build_dir = path.abspath(self.build_temp)
        source_dir = path.dirname(path.abspath(__file__))
        # Keep CMake install artifacts inside the build directory so PEP517
        # isolated builds don't mutate the source tree.
        install_dir = path.join(build_dir, 'cmake-install')

        os.makedirs(build_dir, exist_ok=True)
        os.makedirs(install_dir, exist_ok=True)

        cmake_opts = [
            '-DCMAKE_INSTALL_PREFIX={}'.format(install_dir),
            '-DCMAKE_INSTALL_LIBDIR=lib',
            '-DCMAKE_POSITION_INDEPENDENT_CODE=TRUE',
            '-DHUNTER_ENABLED=OFF',
            '-DMEOWPOW_BUILD_TESTS=OFF',
            '-DMEOWPOW_INSTALL_CMAKE_CONFIG=OFF'
        ]

        # Prefer a deterministic build type for single-config generators.
        if not any(opt.startswith('-DCMAKE_BUILD_TYPE=') for opt in cmake_opts):
            cmake_opts.append('-DCMAKE_BUILD_TYPE=Release')

        generator = os.environ.get('GENERATOR')
        if generator:
            cmake_opts.append('-G{}'.format(generator))

        if not self.skip_cmake_build and not os.environ.get('MEOWPOW_PYTHON_SKIP_BUILD'):
            cmake_cmd = shutil.which('cmake')
            if not cmake_cmd:
                raise CCompilerError(
                    "cmake tool not found but required to build this package")

            r = subprocess.call([cmake_cmd, source_dir] + cmake_opts,
                                cwd=build_dir)
            if r != 0:
                raise CCompilerError(
                    "cmake configuration failed with exit status {}".format(r))
            r = subprocess.call(
                [cmake_cmd, '--build', build_dir, '--target', 'install',
                 '--config', 'Release'])
            if r != 0:
                raise CCompilerError(
                    "cmake build failed with exit status {}".format(r))

        self.library_dirs.append(path.join(install_dir, 'lib'))

        super(build_ext, self).run()


setup(
    name='meowpow',
    version='0.5.2',
    description="C/C++ implementation of Meowpow - the Meowcoin Proof of Work algorithm",
    long_description=path.join(path.dirname(__file__), 'README.md') and open(
        path.join(path.dirname(__file__), 'README.md'), encoding='utf-8'
    ).read(),
    long_description_content_type='text/markdown',
    url='https://github.com/chfast/meowpow',
    author='Pawel Bylica',
    author_email='pawel@ethereum.org',
    license='Apache License, Version 2.0',
    license_files=['LICENSE'],

    package_dir={'': 'bindings/python'},
    packages=['meowpow'],
    cffi_modules=['bindings/python/meowpow/_build.py:ffibuilder'],

    python_requires='>=3.10',
    install_requires=['cffi>=1.12'],

    test_suite='tests.test_meowpow',

    cmdclass={'build_ext': build_ext},

    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: C',
        'Programming Language :: C++',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ],
)


# meowpow: C/C++ implementation of Meowpow, the Meowcoin Proof of Work algorithm.
# Copyright 2019 Pawel Bylica.
# Licensed under the Apache License, Version 2.0.

# The CFFI build script for meowpow library.
# It expects the library is installed in the dist/ directory.
# The installation can be performed by
#
#     cmake . -DCMAKE_INSTALL_PREFIX=dist
#     make
#     make install

from cffi import FFI
import sys

ffibuilder = FFI()

stdlib = []
if sys.platform == 'linux':
    stdlib.append('stdc++')

ffibuilder.set_source(
    "_meowpow",
    r"""
    #include <meowpow/keccak.h>
    #include <meowpow/meowpow.h>
     """,
    include_dirs=['include'],
    libraries=['meowpow', 'keccak'] + stdlib,
)

ffibuilder.cdef("""

union meowpow_hash256
{
    ...;
    char str[32];
};

union meowpow_hash512
{
    ...;
    char str[64];
};

struct meowpow_result
{
    union meowpow_hash256 final_hash;
    union meowpow_hash256 mix_hash;
};


union meowpow_hash256 meowpow_keccak256(const uint8_t* data, size_t size);

union meowpow_hash512 meowpow_keccak512(const uint8_t* data, size_t size);

const struct meowpow_epoch_context* meowpow_get_global_epoch_context(int epoch_number);

struct meowpow_result meowpow_hash(const struct meowpow_epoch_context* context,
    const union meowpow_hash256* header_hash, uint64_t nonce);
    
bool meowpow_verify(const struct meowpow_epoch_context* context,
    const union meowpow_hash256* header_hash, const union meowpow_hash256* mix_hash, uint64_t nonce,
    const union meowpow_hash256* boundary);

union meowpow_hash256 light_verify(const union meowpow_hash256* header_hash, const union meowpow_hash256* mix_hash, uint64_t nonce);                            

""")

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)
